#include <math.h>
#include <stdio.h>
#define DIM 3
#define REAL double
//#define ONLYGRAPE /*(ONLYGRAPE,ONLYHOST,COMPARE)*/
//#define USEGRAPE6 /*(USEHARP1,USEHARP2,USEGRAPE4)*/
/* #define SHAREDTIMESTEP */
#define MAXTIMESTEP 0.125
#define NEMOFORMATOUTPUT /*(NEMOOUTPUTFORMAT,MONGOFORMAT,no-output)*/
#define PI 3.14159265358979323846
#ifdef ONLYHOST
# define NMAX 11000
# define NBMAX 11000
#else
# define NMAX 270000
# define NBMAX 40000
#endif
#ifdef USEGRAPE6
# define NPIPE 128
#else
# define NPIPE 10
#endif
struct previous{
  int flag;
  REAL cod[DIM];
  REAL rc;
  REAL total_m;
  REAL rt;
  REAL wg2;  
  REAL init_ene;
};
#define ONLYHOST_NB

