#define NJMAX 130000
     
struct jlist_t {
/*int index[NJMAX];*/
double mass[NJMAX];   
double x[NJMAX][3];
};   

struct clist_t {
  int next;
/*  int icell;*/
  int key_level;   
  long long int key;
  int ifirst;
  int n;
  REAL mass;
  REAL cm[3];
int gflag; /* 1: group cell 0: upper cell 2: lower cell */
  REAL l_theta;
  REAL length;
  int zeroflag;
  int calcflag;
};
